/*
 * SPDX-FileCopyrightText: 2023 Florian Müllner <fmuellner@gnome.org>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

.system-monitor-stat-section {
  padding: 0 6px;
}

.system-monitor-stat-section-icon {icon-size: 1.08em;}
.system-monitor-stat-section-label {
  min-width: 3.0em;
  font-size: 0.9em;
 }
.system-monitor-placeholder {-st-icon-style: symbolic;}

.panel-button .high-usage {color: #ff7800;}
