/*
 * SPDX-FileCopyrightText: 2013 Florian Müllner <fmuellner@gnome.org>
 * SPDX-FileCopyrightText: 2015 Jakub Steiner <jimmac@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

@import url("stylesheet-dark.css");

.workspace-indicator.name-label {
  box-shadow: inset 0 0 0 100px rgba(34, 34, 38, 0.17) !important;
}
.workspace-indicator.name-label:hover,
.workspace-indicator.name-label:focus {
  box-shadow: inset 0 0 0 100px rgba(34, 34, 38, 0.28) !important;
}
.workspace-indicator.name-label:active {
  box-shadow: inset 0 0 0 100px rgba(34, 34, 38, 0.32) !important;
}
.workspace-indicator.name-label:active:hover {
  box-shadow: inset 0 0 0 100px rgba(34, 34, 38, 0.36) !important;
}

.workspace-indicator .workspace {
  background-color: #ccc;
}

.workspace-indicator .workspace.active {
  border-color: #888;
}

.workspace-indicator-window-preview {
  background-color: #ededed;
  border: 1px solid #ccc;
}

.workspace-indicator-window-preview.active {
  background-color: #f6f5f4;
}

.workspace-indicator-menu .editable-menu-item .icon-button.flat:hover {
  background-color: st-transparentize(black, 90%);
}

.workspace-indicator-menu .editable-menu-item .icon-button.flat:active {
  background-color: st-transparentize(black, 85%);
}

.workspace-indicator-menu .editable-menu-item .icon-button.flat:checked:hover {
  background-color: st-darken(-st-accent-color, 10%);
}

.workspace-indicator-menu .editable-menu-item .icon-button.flat:checked:active {
  background-color: st-darken(-st-accent-color, 15%);
}
