/*
 * SPDX-FileCopyrightText: 2013 Florian Müllner <fmuellner@gnome.org>
 * SPDX-FileCopyrightText: 2015 Jakub Steiner <jimmac@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

@import url("stylesheet-dark.css");
@import url("stylesheet-workspace-switcher-light.css");

#panel.bottom-panel {
    border-top-width: 1px;
    border-bottom-width: 0px;
    height: 2.5em;
  }

  .bottom-panel .window-button > StWidget {
    border-radius: 3px;
    padding: 3px 6px 1px;
    box-shadow: none;
    text-shadow: none;
  }

 .window-button > StWidget {
  color: #000;
}

.window-button:hover > StWidget {
  background-color: st-darken(#eee,5%);
}

.window-button:active > StWidget,
.window-button:focus > StWidget {
  background-color: st-darken(#eee, 10%);
}

.window-button.focused > StWidget {
  background-color: st-darken(#eee,15%);
}

  .window-button.focused:hover > StWidget {
    background-color: st-darken(#eee, 20%);
  }

  .window-button.focused:active > StWidget {
    background-color: st-darken(#eee, 25%);
  }

.window-button.minimized > StWidget {
  color: #aaa;
  background-color: #f9f9f9;
}

.window-button.minimized:active > StWidget {
  color: #aaa;
  background-color: #f9f9f9;
}

.window-button-drag-actor {
  background-color: #ddd;
    border-color: #888;
}

.window-list .placeholder {
  border-color: rgba(0,0,0,0.5);
}
