/*
 * SPDX-FileCopyrightText: 2012 Florian Müllner <fmuellner@gnome.org>
 * SPDX-FileCopyrightText: 2013 Giovanni Campagna <gcampagna@src.gnome.org>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */
@import url("stylesheet-workspace-switcher-dark.css");

.window-list-workspace-indicator.previews {
  -natural-hpadding: 0 !important;
  -minimum-hpadding: 0 !important;
}

.window-list {
  spacing: 2px;
  font-size: 10pt;
}

.bottom-panel {
  background-color: #000000;
  border-top-width: 0px;
  height: 2.45em;
}

.window-button,
.window-button-drag-actor {
  padding: 4px, 3px;
}

.window-button-drag-actor {
  background-color: #444;
  border-radius: 7px;
  border-width: 2px;
  border-color: #fff;
  box-shadow: 0 1px 2px rgba(0,0,0,0.1);
}

.window-button:first-child:ltr {
  padding-left: 2px;
}

.window-button:last-child:rtl {
  padding-right: 2px;
}

.window-button-box {
  spacing: 4px;
}

.window-button > StWidget {
  color: #fff;
  background-color: transparent;
  border-radius: 4px;
  padding: 3px 6px 1px;
  transition: 100ms ease;
}

.window-button > StWidget,
.window-list .placeholder {
  -st-natural-width: 18.75em;
  max-width: 18.75em;
}

.window-list .placeholder {
  border: 1px solid rgba(255,255,255,0.4);
  border-radius: 7px;
  margin: 4px;
}

.window-button:hover > StWidget {
  background-color: #303030;
}

.window-button:active > StWidget,
.window-button:focus > StWidget {
  background-color: st-lighten(#303030, 5%);
}

.window-button.focused > StWidget {
  background-color: #5b5b5b;
}

  .window-button.focused:hover > StWidget {
    background-color: st-lighten(#5b5b5b, 5%);
  }

  .window-button.focused:active > StWidget {
    background-color: st-lighten(#5b5b5b, 10%);
  }

.window-button.minimized > StWidget {
  color: #666;
  background-color: #161616;
}

.window-button.minimized:active > StWidget {
  color: #666;
  background-color: #161616;
}

.window-button-icon {
  width: 24px;
  height: 24px;
}

.window-button-abstract-label {
  background-color: #888;
  border-radius: 99px;
  margin: 6px;
}

.window-button-attention-indicator {
  background-color: -st-accent-color;
  height: 2px;
}

.window-button.minimized .window-button-attention-indicator {
  background-color: st-transparentize(-st-accent-color, 0.4);
}
