/*
 * SPDX-FileCopyrightText: 2020 Florian Müllner <fmuellner@gnome.org>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

.drive-menu-item { spacing: 12px; }

.drive-menu-item .button {
  border-radius: 99px;
  padding: 3px;
  min-height: auto;
}
